--Renoise Keybinds and menus
-------------------------------

TOOL_NAME = "Group Nudge"

--Renoise Keybinds and menus
-------------------------------
renoise.tool():add_keybinding {
  name = "Global:Tools:"..TOOL_NAME,
  invoke = function()main_toggle()end  
}

renoise.tool():add_menu_entry {
  name = "Main Menu:Tools:Ledger`s Scripts:"..TOOL_NAME,
  invoke = function()main_start()end  
}

renoise.tool():add_menu_entry {
  name = "Pattern Editor:Group:"..TOOL_NAME,
  invoke = function()main_start()end  
}

renoise.tool():add_menu_entry {
  name = "Pattern Editor:Group:".."Unmute Groups With Active Tracks",
  invoke = function()unmute_parent_groups_with_active_tracks()end  
}

renoise.tool():add_keybinding {
  name = "Global:Tools:".."Unmute Groups With Active Tracks",
  invoke = function()unmute_parent_groups_with_active_tracks()end  
}

--------------------------------------------------------------------------------
--some basic colors for gui elements
--------------------------------------------------------------------------------
--e.g. For changing vb.views["sample present colour 2"].color when states change
COLOR_GREY = {0x30,0x42,0x42}
COLOR_GREY_MILD = {0x60,0x75,0x75}
COLOR_ORANGE ={0xFF,0x66,0x00}
COLOR_YELLOW = {0xE0,0xE0,0x00}
COLOR_BLUE = {0x50,0x40,0xE0}  
COLOR_RED = {0xEE,0x10,0x10}
COLOR_GREEN = {0x20,0x99,0x20}
COLOR_RED_MILD = {0x90,0x10,0x10}

--Constants holding first index of color-code: Used to identify color in vb.views table i.e.
--if vb.views["button"].color[1] == COLOR_ORANGE_FLAG then --etc
COLOR_GREY_FLAG = COLOR_GREY[1]
COLOR_ORANGE_FLAG = COLOR_ORANGE[1]
COLOR_YELLOW_FLAG = COLOR_YELLOW[1]
COLOR_BLUE_FLAG = COLOR_BLUE[1]  
COLOR_RED_FLAG = COLOR_RED[1]
COLOR_GREEN_FLAG = COLOR_GREEN[1]
COLOR_RED_MILD_FLAG = COLOR_RED_MILD[1]


LUA_COUNTS_FROM_1 = 1

--global variables for gui
local my_dialog = nil 
local vb = nil


--toggle the tool open and closed (keyboard shortcut start-up)
-------------------------------------------------------------
function main_toggle()
----------------------
 --close dialog if it is open
  if (my_dialog and my_dialog.visible) then 
    my_dialog:close()
    --reset global my_dialog
     my_dialog = nil 
  else --run main
    main()
  end
end

--always open/ restart tool (menu entry start-up)
-------------------------------------------------
function main_start()
---------------------
  if (my_dialog and my_dialog.visible) then 
    my_dialog:close()
    --reset global my_dialog
     my_dialog = nil 
  end
  --run main
  main()
end
----------------------------------------------------------


--[[
--for bug testing - rename all tracks (renoise renames tracks when it swaps them and they have their original names)
-------------------------------
renoise.tool():add_keybinding {
  name = "Global:Tools:Rename all tracks",
  invoke = function()rename_tracks()end  
}

function rename_tracks()
  
  local song = renoise.song()
  local grp_count = 1
  
  for i = 1,song.sequencer_track_count do
    local track = song:track(i) 
    local name = track.name
    --grp
    if track.type == renoise.Track.TRACK_TYPE_GROUP then
      track.name = "Grp "..grp_count
      grp_count = grp_count + 1
    else--normal track
      track.name = tostring(i)
    end
  end

end
--------------------------------------------
--]]




--------------------------------------------------------------------------------
--helper function : custom status message --prefixes tool name and adds brackets 
--------------------------------------------------------------------------------
local function status(message)
  renoise.app():show_status(TOOL_NAME.." Tool: ("..message..")")
end

--for extra menu function
--------------------------------------------------
function unmute_parent_groups_with_active_tracks()
--------------------------------------------------

  local song = renoise.song()
  --loop tracks
  for i = 1,#song.tracks do
    --track
    local trk = song:track(i)
    --if an active sequencer track with parents
    if (trk.type == renoise.Track.TRACK_TYPE_SEQUENCER) and 
       (trk.mute_state == renoise.Track.MUTE_STATE_ACTIVE) and 
       (trk.group_parent ~= nil) then 
      --scan all parent groups
      while true do
        --see if track has a parent group (recursive check so not redundant with check above)
        if trk.group_parent ~= nil then
          --update object to next group track
          trk = trk.group_parent
          --unmute parent track
          trk.mute_state = renoise.Track.MUTE_STATE_ACTIVE
        else
          break
        end
      end--while
    end --if
  end--for
end--fn
------

local g_bypass_popup_notifier = false
------------------------------------
function set_popup_to_nearest_group()
------------------------------------
  
  g_bypass_popup_notifier = true 
  
  local song = renoise.song() 
  local target_grp = nil
  local cur_trk = song.selected_track
  --check if selected track is a group
  if cur_trk.type == renoise.Track.TRACK_TYPE_GROUP then
     target_grp = cur_trk
  --if not check if has a parent group  
  elseif cur_trk.group_parent ~= nil then
    target_grp = cur_trk.group_parent
  end

  --if target_grp found then update popup to match
  if target_grp ~= nil then
    local grp_table = get_current_groups().group_obj
    for i = 1, #grp_table do
       if rawequal(grp_table[i],target_grp) == true then
         vb.views["target group track"].value = i
         break
       end
    end
  end
  g_bypass_popup_notifier = false 
end

--function solos tracks one by one within a group
------------------------------------------------
function cycle_solo_within_selected_group(grp_obj,grp_idx)
-----------------------------------------------
  
  local song = renoise.song()
  local selected_track = song.selected_track
  local parent_group = grp_obj
  
  --return if we are not on a group member track
  if parent_group == nil then
    status("Track Not In A Group-Track")
    return
  end
  
  --loop to make sure all parent groups are unmuted
   --scan all parent groups
  local trk = song.selected_track
  --make sure parent group is unmuted as it might be the selected track
  parent_group.mute_state = renoise.Track.MUTE_STATE_ACTIVE
  while true do
    --see if track has a parent group (recursive check so not redundant with check above)
    if trk.group_parent ~= nil then
      --update object to next group track
      trk = trk.group_parent
      --unmute parent track
      trk.mute_state = renoise.Track.MUTE_STATE_ACTIVE
    else
      break
    end
  end

  
   local unmuted_track = nil
   local end_track = grp_idx - 1
   local start_track = end_track - #parent_group.members + 1
   --loop backwards through group members to find the last (most rightward) active member track
   for i = end_track,start_track,-1 do
     local trk = song:track(i) 
     if (trk.mute_state == renoise.Track.MUTE_STATE_ACTIVE) and (unmuted_track == nil) then
       unmuted_track = i
     end
      --track been checked so mute it
      trk:mute()
   end
   
   --update start_track, if the first track/s are in a sub group we want to move start track 
   --up until it is only one levl down from main group
   for i = start_track,end_track do
     if rawequal(song:track(i).group_parent,parent_group) == true then
       start_track = i
       break
     end
   end
      
   if unmuted_track == nil then --default to end track so that start_track will be the next unmuted track after calculation below
     unmuted_track = end_track
   end
  

  -----------
  
  ------------
  
  local next_track = unmuted_track + 1
  
  while true do
    if (rawequal(song:track(next_track).group_parent,parent_group) == false) and (next_track < end_track) then
      next_track = next_track + 1
    else
      break
    end
  end
   
  --set new active track
  local active_track = nil
  if (next_track <= end_track) then
     active_track = song:track(next_track)
     active_track.mute_state = renoise.Track.MUTE_STATE_ACTIVE
  else
    if start_track == unmuted_track then
      status("Only One Member In This Group")
    end        
    active_track = song:track(start_track)
    active_track.mute_state = renoise.Track.MUTE_STATE_ACTIVE 
  end
  

   --if the active track is a sub-group track, make sure all its members are unmuted
   if active_track.type == renoise.Track.TRACK_TYPE_GROUP then
     for i = 1,#active_track.members do
      -- print( active_track.members[i].name)
       active_track.members[i].mute_state = renoise.Track.MUTE_STATE_ACTIVE
     end
   end
  
 
end
------------------------------------------------------------------------
------------------------------------------------------------------------





--gets a table containing all the current group tracks in the song
--and their track indexes.  For populating the popup.
-----------------------------
function get_current_groups()
-----------------------------
  
  --table to hold name, index and pointer ref of group tracks (functions return value)
  local grp_trks = {}
  grp_trks.names = {}
  grp_trks.index = {}
  grp_trks.group_obj = {} --obj (pointer ref)
  
  --song object
  local song = renoise.song()
  local counter = 1
  --loop tracks in song (reversed)
  for i = #song.tracks,1,-1 do
    --populate table when gropu track found
    if song:track(i).type == renoise.Track.TRACK_TYPE_GROUP then
      grp_trks.names[counter] = song:track(i).name
      grp_trks.index[counter] = i
      grp_trks.group_obj[counter] = song:track(i) 
      counter = counter + 1 
    end
  end
  --return table
  return grp_trks
end
----------------------------
----------------------------

--returns the renoise track-index of the first track in the specified group
--------------------------------------------
function get_index_of_first_track_in_group(group_index)
--------------------------------------------
  
  local song = renoise.song()
  local group_track = song:track(group_index)
  --if group has no members then return nil
  if #group_track.members == 0 then
    return nil
  end
  --get the index of the tleft hand group
  local left_of_grp_idx = (group_index - #group_track.members) 
  --returns index
  return left_of_grp_idx
end
---------------------------
---------------------------

--extend left border of group leftwards by one track
---------------------------------------  
function extend_group_left(group_index)  
---------------------------------------
  
  local song = renoise.song()
  local group_track = song:track(group_index)

  --get the index of the track to the left of the group.  This is the track we want to add to the group
  local left_of_grp_idx = (group_index - #group_track.members - 1) 
  
  --if left_of_grp_idx == 0 then there are no more tracks to the left so return early 
  if left_of_grp_idx == 0 then 
    return
  end
  
  --if group has no members already, then it is a simple add track to group operation and
  --no re-organisation is needed
  if #group_track.members == 0 then
    song:add_track_to_group(left_of_grp_idx,group_index)
    return
  end
  
  --1) Add 2 dummy tracks into group
  ----------------------------------
  --NOTE We need to insert 2 dummy tracks to account for bugs with the renoise swap tracks behaviour within groups
  --see:  http://forum.renoise.com/index.php/topic/51960-lua-api-bug-swap-tracks/
  
  --set member_idx to the first member track of the group
  local member_idx = left_of_grp_idx + 1

  --loop group members to find insert point for dummy tracks (after first legitimate member; i.e. not a track in a further sub-group)
  for i = 1,#group_track.members do
    --check if member is not in a sub group by checking its parent
    if rawequal(song:track(member_idx).group_parent,group_track) == true then
      break
    end
    --if here then member was in a sub- group so increment member_idx until we get to a legitimate track
    member_idx = member_idx + 1
  end
  
  --again due to insert_track behaviour we need to insert the dummy tracks after the first present and legitimate member of the group
  --so + 1
  local insert_point = member_idx + 1
  
  --insert 2 dummy tracks at member_idx 
  local dummy_trk_2 = song:insert_track_at(insert_point)
  local dummy_trk_1 = song:insert_track_at(insert_point)
  
  --name the tracks for bug hunting i.e. if they are not deleted we will see them in pattern ed
  dummy_trk_1.name = "Dummy Track 1"
  dummy_trk_2.name = "Dummy Track 2"
  
  --2) Swap Dummy tracks to members 1 and 2 of group (this works re. the swap bug as we are not swapping two groups)
  song:swap_tracks_at(member_idx,member_idx+1)
  song:swap_tracks_at(member_idx+1,member_idx+2)
  
  --3) Swap dummy member 2 with target track outside group (left_of_grp_idx was found near beginning of function)
  --note: if we were to swap dummy member 1 then both tracks get ejected from group
  song:swap_tracks_at(left_of_grp_idx,(left_of_grp_idx + 2))
 
  --4) delete dummy tracks by looping objects 
  --(as we are deleting tracks it is safer to loop and find the objects specifically or
  -- we might lose wanted song tracks if there was an arithmetic mistake if a bug is present)
  for i = 1,#song.tracks do
    if rawequal(song:track(i),dummy_trk_1) then
      --delete dummy 1
      renoise.song():delete_track_at(i)
      break
    end
  end
  
  for i = 1,#song.tracks do
    if rawequal(song:track(i),dummy_trk_2) then
    --delete dummy 2
      renoise.song():delete_track_at(i)
      break
    end
  end
end 



--return table of all non-empty note object in a track 
--nil if no notes  -- print format will look like: >> [200] =>  B-50A......0000
--------------------------------------------------
local function get_all_notes_in_track(trk_idx)
--------------------------------------------------
  local song = renoise.song()
  --table to hold all notes
  local note_tab = {}
  --track object
  local trk = song:track(trk_idx)
  --make sur seq track
  if trk.type == renoise.Track.TRACK_TYPE_SEQUENCER then
    --loop patterns
    for pat = 1,#song.patterns do
      local pat_trk = song:pattern(pat):track(trk_idx)
      if pat_trk.is_empty == false then
        --loop lines in pattern
        for line = 1,#pat_trk.lines do
          local cur_line = pat_trk:line(line)
          if cur_line.is_empty == false then
            --loop columns of line
            for col = 1,12 do
              local note = cur_line:note_column(col)
              if note.is_empty == false then
                if note.note_string ~= "---" and note.note_string ~= "OFF" then
                  table.insert(note_tab,note)
                end
              end  
            end
          end
        end
      end
    end
  end
  if #note_tab == 0 then
    return nil
  else
    return note_tab
  end
end




------------------------------------
------------------------------------
function main()
------------------------------------
------------------------------------

  --renoise song object
  local song = renoise.song()
  

  --set globals
  -------------
  --assign global vb to viewbuilder
  vb = renoise.ViewBuilder()
  
  ----------------------------------------------
  --GUI
  ----------------------------------------------     

  --variables that will be added to
  --dialog_content:add_child(send_row)
  local my_first_row = vb:row {}
  -------------------------------
  local my_first_element = nil
  

    --GUI
    my_first_element = vb:column{
                        margin = 4,
                        
                        vb:row{ 
                         vb:text{
                          text = "Target Group",
                          },
                         vb:text{
                           width = 65,
                         },
                         --[[
                         vb:button{
                          color = COLOR_GREY,
                          id = "select button",
                          notifier = function() 
                                        set_popup_to_nearest_group()
                                     end
                           },
                         --]]
                        },
                        
                        vb:row{ --row 1
                         margin = 2,
                         vb:popup{                               
                          id = "target group track",
                          items = get_current_groups().names,
                          width = 150,
                          notifier = function(value)
                                       
                                       if g_bypass_popup_notifier == true then
                                         return
                                       end
                                       
                                       local song = renoise.song()
                                       song.selected_track_index = get_current_groups().index[value]
                                      
                                      -- vb.views["grp fx off"].color = COLOR_GREY
                                      -- vb.views["grp fx on"].color = COLOR_GREY                                  
                                     end
                          }, 
                         },
                         
                         vb:row{ --row 2
                           margin = 2,
                            vb:button{
                             height = 25,
                             text = "<<",
                             color = COLOR_GREY,
                             id = "nudge left",
                             notifier = function()
                                         
                                         local song = renoise.song()
                                         
                                         --get selected_track index so we can preserve it
                                         local sel_idx = song.selected_track_index
                                         
                                         --get target group track index from popup
                                         local popup_selected_group_track = vb.views["target group track"].value
                                         local group_index = get_current_groups().index[popup_selected_group_track]
                                         --get the track index in renoise of the first member track in the group
                                         local first_of_grp = get_index_of_first_track_in_group(group_index)
                                         
                                         --we have reached track 1 so no more tracks to be added, so return here
                                         if first_of_grp == 1 then
                                           return
                                         end
                                         --no member tracks present so set first-of_grp to group_index
                                         if first_of_grp == nil then
                                           first_of_grp = group_index
                                         end
                                            
                                          extend_group_left(group_index)                                      
                                         --add track to the left of group (first_of_grp - 1), into the group
                                         -- add_track_as_first_member_of_group((first_of_grp - 1),group_index)
                                        
                                         --re selected original track
                                         song.selected_track_index = sel_idx
                                      
                                       end
                           
                           
                           },
                           
                           vb:button{
                             height = 25,
                             id = "nudge right",
                             text = ">>",
                             color = COLOR_GREY,
                             notifier = function()
                                         
                                         

                                         local song = renoise.song()
                                         --get selected_track index so we can preserve it
                                         local sel_idx = song.selected_track_index
                                         
                                         --get target group track index from popup
                                         local popup_selected_group_track = vb.views["target group track"].value
                                         local group_index = get_current_groups().index[popup_selected_group_track]
                                         --selected_group object
                                         local selected_group = song:track(group_index)
                                         
                                         --is the chosen group a sub-group --gets group track or nil (and )
                                         local parent_group = song:track(group_index).group_parent

                                         --get the track index in renoise of the first member track in the group
                                         local first_of_grp = get_index_of_first_track_in_group(group_index)
                                         
                                         if first_of_grp == nil then
                                           return
                                         end
                                         
                                         --due to sub grouping the first track present in the group may not be the track to remove
                                         local track_to_remove = first_of_grp
                                         --loop rightwards through tracks until a track is reached that has the main group as it`s parent is found
                                         --if already true then no looping is done.
                                         while (rawequal(song:track(track_to_remove).group_parent,song:track(group_index)) ~= true) do
                                           track_to_remove =  track_to_remove + 1
                                         end

                                         song:remove_track_from_group(track_to_remove)
                                         
                                       
                                         
                                         --if there is a parent group we need to add the removed track back into the parent group
                                         --otherwise parent group gets shrunk aswell
                                         if (parent_group ~= nil)  then
                                           
                                           local parent_group_index = nil
                                           --get the parent_group index by looping renoise track objects
                                           for i = 1,song.sequencer_track_count do
                                             if rawequal(parent_group, song:track(i)) == true then
                                               parent_group_index = i
                                               --print(parent_group_index)
                                               break
                                             end
                                           end
                                           
                                           local selected_group_index = nil
                                           --get the parent_group index by looping renoise track objects
                                           for i = 1,song.sequencer_track_count do
                                             if rawequal(selected_group, song:track(i)) == true then
                                               selected_group_index = i
                                               --print(parent_group_index)
                                               break
                                             end
                                           end
                                           
                                           
                                           local parent_group_first_member_pos = parent_group_index - #parent_group.members - 1
                                           local selected_group_first_member_pos = selected_group_index - #selected_group.members - 1
                                           
                                           if parent_group_first_member_pos == selected_group_first_member_pos then
                                             extend_group_left(parent_group_index)   
                                           end
                                         
                                         end
                                          --re selected original track
                                          song.selected_track_index = sel_idx
                                        end
                             },
                             
                             
                             
                             vb:text{
                              text = "Nudge Left Border",
                             },
                            
                            },
                            
                            vb:row{ --spacer
                              margin = 2
                            },
                           --[[ 
                            vb:row{
                             margin = 2,
                             vb:text{ 
                              text = "Group FX"
                             }
                            },
                            
                            vb:row{ --row 3
                             margin = 2,
                             vb:button{
                               color = COLOR_GREY,
                               id = "grp fx on",
                               text = "Grp FX ON",
                               notifier = function()
                                            local song = renoise.song()
                                            --get popup value for target group
                                            local selected_target_grp = vb.views["target group track"].value
                                            --get target groups index passing popup value
                                            local tar_grp_index = get_current_groups().index[selected_target_grp]
                                            --get target group track object
                                            local tar_grp_track = song:track(tar_grp_index)
                                            
                                            --enable  devices on group  track
                                            for i = 2,#tar_grp_track.devices do
                                              tar_grp_track:device(i).is_active = true
                                            end
                                            
                                            vb.views["grp fx on"].color = COLOR_ORANGE
                                            vb.views["grp fx off"].color = COLOR_GREY
                                            
                                          end
                                     
                             }, --button
                             vb:button{
                               color = COLOR_GREY,
                               id = "grp fx off",
                               text = "Grp FX OFF",
                               notifier = function()
                                            local song = renoise.song()
                                            --get popup value for target group
                                            local selected_target_grp = vb.views["target group track"].value
                                            --get target groups index passing popup value
                                            local tar_grp_index = get_current_groups().index[selected_target_grp]
                                            --get target group track object
                                            local tar_grp_track = song:track(tar_grp_index)
                                            
                                            --disable devices on group  track
                                            for i = 2,#tar_grp_track.devices do
                                              tar_grp_track:device(i).is_active = false
                                            end
                                            
                                            vb.views["grp fx off"].color = COLOR_ORANGE
                                            vb.views["grp fx on"].color = COLOR_GREY
                                          end
                                     
                             } --button
                             
                           },--row 3--]]
                           
                           vb:row{
                             margin = 2,
                             vb:text{ 
                              text = "Solo Grp Members"
                            }
                           
                           },
                           
                           
                           
                         vb:row{
                           margin = 2,
                           vb:button{
                             text = "Solo Each",
                             id = "solo each btn",
                             color = COLOR_GREY,
                             notifier = function()
                                         --song
                                         local song = renoise.song()
                                         --get popup index for selected group track
                                         local popup_sel = vb.views["target group track"].value
                                         --get group index
                                         local sel_grp_idx = get_current_groups().index[popup_sel]
                                         --get group object pointer
                                         local grp_obj = song:track(sel_grp_idx)
                                         --call function to cycle solo through tracks within the group
                                         cycle_solo_within_selected_group(grp_obj,sel_grp_idx)
                                        end,
                                         
                           },
                           
                          vb:button{
                             text = "Clear",
                             id = "clear btn",
                             color = COLOR_GREY,
                             notifier = function()
                                          --song
                                          local song = renoise.song()
                                          --get popup index for selected group track
                                          local popup_sel = vb.views["target group track"].value
                                          --get group index
                                          local sel_grp_idx = get_current_groups().index[popup_sel]
                                          --get group object pointer
                                          local grp_obj = song:track(sel_grp_idx)
                                          --get index of first group member
                                          local first_mem_idx = sel_grp_idx - #grp_obj.members
                                          --loop range to unmute all tracks
                                          for i = first_mem_idx, sel_grp_idx do
                                            song:track(i).mute_state = renoise.Track.MUTE_STATE_ACTIVE
                                          end
                                        end
                                         
                           }, 
                         },-- row
                         
                         vb:row{
                          vb:text{
                            text = "Tr. Group         Tr. Track"
                          
                          }
                         
                         
                         },--row
                         vb:row{
                           margin = 2,
                           vb:button{ --transpose selected group -
                             color = COLOR_GREY,
                             width = 30,
                             text = "-",
                             id = "tr grp btn -",
                             notifier = function()
                                          --song
                                          local song = renoise.song()
                                          --get popup index for selected group track
                                          local popup_sel = vb.views["target group track"].value
                                          --get group index
                                          local sel_grp_idx = get_current_groups().index[popup_sel]
                                          --get group object pointer
                                          local grp_obj = song:track(sel_grp_idx)
                                          --first member idx
                                          local first_idx = sel_grp_idx - #grp_obj.members
                                          local last_idx = sel_grp_idx - 1
    
                                          --use get_all_notes_in_track(track_index) function
                                          --loop all group members
                                          for i = first_idx,last_idx do
                                            local notes = get_all_notes_in_track(i)
                                            if notes ~= nil then
                                              --loop all notes in track
                                              for j = 1,#notes do
                                                local note = notes[j] 
                                                 if note.note_value > 0 then
                                                  note.note_value = note.note_value - 1
                                                end
                                              end
                                            end
                                          end
                                        end
                             
                           

                           },
                           
                           vb:button{ --transpose selected group +
                             color = COLOR_GREY,
                             width = 30,
                             text = "+",
                             id = "tr grp btn +",
                             notifier = function()
                                          --song
                                          local song = renoise.song()
                                          --get popup index for selected group track
                                          local popup_sel = vb.views["target group track"].value
                                          --get group index
                                          local sel_grp_idx = get_current_groups().index[popup_sel]
                                          --get group object pointer
                                          local grp_obj = song:track(sel_grp_idx)
                                          --first member idx
                                          local first_idx = sel_grp_idx - #grp_obj.members
                                          local last_idx = sel_grp_idx - 1
                                         
                                          --use get_all_notes_in_track(track_index) function
                                          --loop all group members
                                          for i = first_idx,last_idx do
                                            local notes = get_all_notes_in_track(i)
                                            if notes ~= nil then
                                              --loop all notes in track
                                              for j = 1,#notes do
                                                local note = notes[j] 
                                                 if note.note_value < 119 then
                                                  note.note_value = note.note_value + 1
                                                end
                                              end
                                            end
                                          end
                                        end                           
                           
                           
                           },
                           
                           vb:text{                           
                             width = 9,
                           },
                           
                           vb:button{ --transpose selected track -
                             color = COLOR_GREY,
                             width = 30,
                             text = "-",
                             notifier = function()
                                          --song
                                          local song = renoise.song()

                                          --use get_all_notes_in_track(track_index) function
                                          local trx_idx = song.selected_track_index
                                          local notes = get_all_notes_in_track(trx_idx)
                                          if notes ~= nil then
                                            --loop all notes in track
                                            for i = 1,#notes do
                                              local note = notes[i] 
                                              if note.note_value > 0 then
                                                note.note_value = note.note_value - 1
                                              end
                                            end
                                          end
                                        
                                        end
                             
                           
                           
                           },
                           
                           vb:button{  --transpose selected track +
                             color = COLOR_GREY,
                             width = 30,
                             text = "+",
                             notifier = function()
                                          --song
                                          local song = renoise.song()

                                          --use get_all_notes_in_track(track_index) function
                                          local trx_idx = song.selected_track_index
                                          local notes = get_all_notes_in_track(trx_idx)
                                          if notes ~= nil then
                                            --loop all notes in track
                                            for i = 1,#notes do
                                              local note = notes[i] 
                                              if note.note_value < 119 then
                                                note.note_value = note.note_value + 1
                                              end
                                            end
                                          end

                                        end
                             
                           
                           
                           },
                          --[[ 
                           vb:checkbox{
                             id = "octave cb",
                           
                           },
                           
                          --]] 
                           
                           
                         
                         }--row

                        }--column

    
    my_first_row:add_child(my_first_element) 
  

  
  --------------------------------------------------------
  --------------------------------------------------------
  --dialog content will contain all of gui; passed to renoise.app():show_custom_dialog()
  local dialog_content = vb:column{}
  dialog_content:add_child(my_first_row)


  
  --------------
  --key Handler
  --------------
  local function my_keyhandler_func(dialog,key)
     --toggle lock focus hack, allows pattern ed to get key input
     renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
     renoise.app().window.lock_keyboard_focus = not renoise.app().window.lock_keyboard_focus
     return key
  end 
  
 
  ------------------------------------------------------------------------------------------
  ------------------------------------------------------------------------------------------
  --Script dialog
  my_dialog = renoise.app():show_custom_dialog(
      TOOL_NAME, dialog_content,my_keyhandler_func)

  
  --add timer to fire once every 50ms
  if not renoise.tool():has_timer(timer) then
    renoise.tool():add_timer(timer,80)
  end
  
  
  --------------------------------------------    
  --close dialog function ON NEW SONG
  --------------------------------------------
  local function closer(d)
    --close dialog if exists and is open     
    if (d ~= nil) and (d.visible == true) then
      d:close()
     -- remove_notifiers()
    end
    --reset global my_dialog
     my_dialog = nil 
  end
  -- notifier to close dialog on load new song
  renoise.tool().app_release_document_observable:add_notifier(closer,my_dialog)
  -------------------------------------------------------------------------------
  
  --setup
  --set the popup to the current group or parent group if tool is opened there
  -------------------------------------
  set_popup_to_nearest_group()


  
  --first run of timer
  timer()
end


--local g_num_groups = nil
----------------
--timer function
----------------
function timer()
  
  local song = renoise.song()
  
  --refresh popup group list
  local name_table = get_current_groups().names
  vb.views["target group track"].items = name_table
  
  
  if #name_table ~= 0 then
   --enable elements
  -- vb.views["grp fx off"].active = true
  -- vb.views["grp fx on"].active = true
   vb.views["nudge right"].active = true
   vb.views["nudge left"].active = true
   vb.views["solo each btn"].active = true
   vb.views["clear btn"].active = true
   vb.views["tr grp btn +"].active = true
   vb.views["tr grp btn -"].active = true
   
  
  set_popup_to_nearest_group()

  --if no groups in song then disable gui elements
  ------------------------------------------------
  else-- #name_table == 0 so:
   -- vb.views["grp fx off"].active = false
    --vb.views["grp fx on"].active = false
   -- vb.views["grp fx off"].color = COLOR_GREY
   -- vb.views["grp fx on"].color = COLOR_GREY
    vb.views["nudge right"].active = false
    vb.views["nudge left"].active = false
    vb.views["solo each btn"].active = false
    vb.views["clear btn"].active = false
    vb.views["tr grp btn +"].active = false
    vb.views["tr grp btn -"].active = false
  end
 
  --remove timer when GUI is closed
  if(my_dialog == nil) or (my_dialog.visible == false) then
    if renoise.tool():has_timer(timer) then
      renoise.tool():remove_timer(timer)
     -- g_num_groups = nil
    end
  end
end








